#!/usr/bin/env python
"""
crackfortran --- read fortran (77,90) code and extract declaration information.
    Usage is explained in the comment block below.

Copyright 1999,2000 Pearu Peterson all rights reserved,
Pearu Peterson <pearu@ioc.ee>          
Permission to use, modify, and distribute this software is given under the
terms of the LGPL.  See http://www.fsf.org

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
$Date: 2000/09/11 20:34:41 $
Pearu Peterson
"""
__version__ = "$Revision: 1.52 $"[10:-1]

f2py_version='See `f2py -v`'

"""
 Usage of crackfortran:
 ======================
 Command line keys: -quite,-verbose,-fix,-f77,-f90,-show,-h <pyffilename>
                    -m <module name for f77 routines>,--ignore-contains
 Functions: crackfortran, crack2fortran
 The following Fortran statements/constructions are supported
 (or will be if needed):
    block data,byte,call,character,common,complex,contains,data,
    dimension,double complex,double precision,end,external,function,
    implicit,integer,intent,interface,intrinsic,
    logical,module,optional,parameter,private,public,
    program,real,(sequence?),subroutine,type,use,virtual,
    include
 Note: 'virtual' is mapped to 'dimension'.
 Note: 'implicit integer (z) static (z)' is 'implicit static (z)' (this is minor bug).
 Note: code after 'contains' will be ignored until its scope ends.
 Note: 'common' statement is extended: dimensions are moved to variable definitions
 Note: f2py directive: <commentchar>f2py<line> is read as <line>

 Usage:
   `postlist=crackfortran(files,funcs)`
   `postlist` contains declaration information read from the list of files `files`.
   `crack2fortran(postlist)` returns a fortran code to be saved to pyf-file

   `postlist` has the following structure:
  *** it is a list of dictionaries containing `blocks':
      B = {'block','body','vars'[,'name','prefix','args','result',
           'implicit','externals','interfaced','common','sortvars',
           'commonvars','note']}
      B['block'] = 'interface' | 'function' | 'subroutine' | 'module' |
                   'program' | 'block data' | 'type'
      B['body'] --- list containing `subblocks' with the same structure as `blocks'
      B['vars'] --- dictionary of variable definitions
      B['sortvars'] --- dictionary of variable definitions sorted by dependence (independent first)
      B['name'] --- name of the block (not if B['block']=='interface')
      B['prefix'] --- prefix string (only if B['block']=='function')
      B['args'] --- list of argument names if B['block']== 'function' | 'subroutine'
      B['result'] --- name of the return value (only if B['block']=='function')
      B['implicit'] --- dictionary {'a':<variable definition>,'b':...} | None
      B['externals'] --- list of variables being external
      B['interfaced'] --- list of variables being external and defined
      B['common'] --- dictionary of common blocks (list of objects)
      B['commonvars'] --- list of variables used in common blocks (dimensions are moved to variable definitions)
      B['from'] --- string showing the 'parents' of the current block
      B['use'] --- dictionary of modules used in current block:
          {<modulename>:{['only':<0|1>],['map':{<local_name1>:<use_name1>,...}]}}
      B['note'] --- list of LaTeX comments on the block
  *** Variable definition is a dictionary
      D = B['vars'][<variable name>] =
      {'typespec'[,'attrspec','kindselector','charselector','=','typename']}
      D['typespec'] = 'byte' | 'character' | 'complex' | 'double complex' |
                      'double precision' | 'integer' | 'logical' | 'real' | 'type'
      D['attrspec'] --- list of attributes (e.g. 'dimension(<arrayspec>)',
                        'external','intent(in|out|inout|hide)','optional',
                        'required', etc)
      K = D['kindselector'] = {['*','kind']} (only if D['typespec'] =
                          'complex' | 'integer' | 'logical' | 'real' )
      C = D['charselector'] = {['*','len','kind']}
                              (only if D['typespec']=='character')
      D['='] --- initialization expression string
      D['typename'] --- name of the type if D['typespec']=='type'
      D['dimension'] --- list of dimension bounds
      D['intent'] --- list of intent specifications
      D['depend'] --- list of variable names on which current variable depends on
      D['check'] --- list of C-expressions; if C-expr returns zero, exception is raised
      D['note'] --- list of LaTeX comments on the variable
  *** Meaning of kind/char selectors (few examples):
      D['typespec>']*K['*']
      D['typespec'](kind=K['kind'])
      character*C['*']
      character(len=C['len'],kind=C['kind'])
      (see also fortran type declaration statement formats below)

 Fortran 90 type declaration statement format (F77 is subset of F90)
====================================================================
 (Main source: IBM XL Fortran 5.1 Language Reference Manual)
 type declaration = <typespec> [[<attrspec>]::] <entitydecl>
 <typespec> = byte                          |
              character[<charselector>]     |
              complex[<kindselector>]       |
              double complex                |
              double precision              |
              integer[<kindselector>]       |
              logical[<kindselector>]       |
              real[<kindselector>]          |
              type(<typename>)
 <charselector> = * <charlen>               |
              ([len=]<len>[,[kind=]<kind>]) |
              (kind=<kind>[,len=<len>])
 <kindselector> = * <intlen>                |
              ([kind=]<kind>)
 <attrspec> = comma separated list of attributes.
              Only the following attributes are used in
              building up the interface:
                 external
                 (parameter --- affects '=' key)
                 optional
                 intent
              Other attributes are ignored.
 <intentspec> = in | out | inout
 <arrayspec> = comma separated list of dimension bounds.
 <entitydecl> = <name> [[*<charlen>][(<arrayspec>)] | [(<arrayspec>)]*<charlen>]
                       [/<init_expr>/ | =<init_expr>] [,<entitydecl>]

 In addition, the following attributes are used: check,depend,note

 TODO:
     * Apply 'parameter' attribute (e.g. 'integer parameter :: i=2' 'real x(i)'
                                    -> 'real x(2)')
     The above may be solved by creating appropriate preprocessor program, for example.
"""
#
import sys,string,fileinput,re,pprint,os,copy
from auxfuncs import *

# Global flags:
strictf77=1          # Ignore `!' comments unless line[0]=='!'
sourcecodeform='fix' # 'fix','free'
quite=0              # Be verbose if 0
verbose=0            # Be "very" verbose if 1
tabchar=4*' '
pyffilename=''
f77modulename=''
skipemptyends=0      # for old F77 programs without 'program' statement
ignorecontains=1
dolowercase=1
debug=[]
# Some helper functions
def show(o,f=0):pprint.pprint(o)
errmess=sys.stderr.write
def outmess(line,flag=1):
    global filepositiontext
    if not quite:
        if flag:sys.stdout.write(filepositiontext)
        sys.stdout.write(line)
re._MAXCACHE=50
defaultimplicitrules={}
for c in "abcdefghopqrstuvwxyz$_": defaultimplicitrules[c]={'typespec':'real'}
for c in "ijklmn": defaultimplicitrules[c]={'typespec':'integer'}
badnames={}
invbadnames={}
for n in ['int','double','float','char','short','long','void','case','while',
          'return','signed','unsigned','if','for','typedef','sizeof','union',
          'struct','static','register','new','break','do','goto','switch',
          'continue','else','inline','extern','delete','const','auto',
          'len','rank','shape','index','slen','size','_i',
          'string','complex_double','float_double','stdin','stderr','stdout']:
    badnames[n]=n+'_v'
    invbadnames[n+'_v']=n
def rmbadname1(name):
    if badnames.has_key(name):
        errmess('rmbadname1: Replacing "%s" with "%s".\n'%(name,badnames[name]))
        return badnames[name]
    return name
def rmbadname(names): return map(rmbadname1,names)
def getextension(name):
    i=string.rfind(name,'.')
    if i==-1: return ''
    if '\\' in name[i:]: return ''
    return name[i+1:]
####### Read fortran (77,90) code
gotnextfile=1
filepositiontext=''
currentfilename=''
def readfortrancode(ffile,dowithline=show,istop=1):
    """
    Read fortran codes from files and
     1) Get rid of comments, line continuations, and empty lines; lower cases.
     2) Call dowithline(line) on every line.
     3) Recursively call itself when statement \"include '<filename>'\" is met.
    """
    global gotnextfile,filepositiontext,currentfilename,sourcecodeform,strictf77,\
           beginpattern,quite,verbose,dolowercase
    if not istop:
        saveglobals=gotnextfile,filepositiontext,currentfilename,sourcecodeform,strictf77,\
           beginpattern,quite,verbose,dolowercase
    if ffile==[]: return
    cont=0
    finalline=''
    ll=''
    commentline=re.compile(r'(?P<line>([^"]*"[^"]*"[^"!]*|[^\']*\'[^\']*\'[^\'!]*|[^!]*))!+(?P<rest>.*)')
    includeline=re.compile(r'\s*(include|INCLUDE)\s*\'(?P<name>.*?)\'')
    cont1=re.compile(r'(?P<line>.*)&\s*\Z')
    cont2=re.compile(r'(\s*&|)(?P<line>.*)')
    if istop: dowithline('',-1)
    ll=''
    spacedigits=[' ']+map(str,range(10))
    filepositiontext=''
    fin=fileinput.FileInput(ffile)
    for l in fin:
        if fin.isfirstline():
            filepositiontext=''
            currentfilename=fin.filename()
            outmess('\tReading file %s\n'%(`currentfilename`))
            gotnextfile=1
            l1=l
            ext = getextension(currentfilename) # fortran form (f77,fix,free) from extension
            if ext in ['f77','F77','for','FOR']: strictf77=1
            else: strictf77=0
            if ext in ['f','F','f77','F77','for','FOR']:
                sourcecodeform='fix'
            elif ext in ['f90','F90','pyf','m']:
                sourcecodeform='free'
            if re.match(r'\A!%.*?\bf77\b',l):sourcecodeform='fix';strictf77=1
            elif re.match(r'\A!%.*?\bf90\b',l):sourcecodeform='free';strictf77=0
            if re.match(r'\A!%.*?\bfix\b',l):sourcecodeform='fix'
            if strictf77: beginpattern=beginpattern77
            else: beginpattern=beginpattern90
            if re.match(r'\A!%.*?\bverbose\b',l):verbose=1
            elif re.match(r'\A!%.*?\bquite\b',l):quite=1
        l=string.expandtabs(l)
        while not l=='':                       # Get rid of newline characters
            if l[-1] not in "\n\r\f": break
            l=l[:-1]
        if not strictf77:
            r=commentline.match(l)
            if r:
                l=r.group('line')# Strip comments starting with `!'
                rl=r.group('rest')
                if string.lower(rl)=='f2py': # f2py directive
                    r=commentline.match(rl[4:])
                    if r:
                        l=l+r('line')
        if string.strip(l)=='': cont=0;continue              # Skip empty line
        if sourcecodeform=='fix':
            if l[0] in ['*','c','!','C','#']:
                if string.lower(l[1:5])=='f2py': # f2py directive
                    l='     '+l[5:]
                else:
                    cont=0; continue # Skip comment line
            if not l[0] in spacedigits:
                raise 'readfortrancode: Found non-(space,digit) char in the first column.\n\tAre you sure that this code is in fix form?\n\tline=%s'%`l`
            if not l[5]==' ':                # Continuation of a previous line
                ll=ll+l[6:]
                finalline=''
            else:
                if dolowercase: finalline=string.lower(ll)
                else: finalline=ll
                ll=l
        elif sourcecodeform=='free':
            r=cont1.match(l)
            if r: l=r.group('line') # Continuation follows ..
            if cont:
                ll=ll+cont2.match(l).group('line')
                finalline=''
            else:
                if dolowercase: finalline=string.lower(ll)
                else: finalline=ll
                ll=l
            cont=(r is not None)
        else:
            raise ValueError,"Flag sourcecodeform must be either 'fix' or 'free': %s"%`sourcecodeform`
        filepositiontext='Line #%d in %s:"%s"\n\t' % (fin.filelineno()-1,currentfilename,l1)
        m=includeline.match(finalline)
        if m:
            fn=m.group('name')
            if os.path.isfile(fn):
                readfortrancode(fn,dowithline=dowithline,istop=0)
            else:
                outmess('readfortrancode: %s is not an existing file. Ignoring.\n'%(`fn`))
        else:
            dowithline(finalline)
        l1=ll
    if dolowercase: finalline=string.lower(ll)
    else: finalline=ll
    filepositiontext='Line #%d in %s:"%s"\n\t' % (fin.filelineno()-1,currentfilename,l1)
    m=includeline.match(finalline)
    if m:
        fn=m.group('name')
        if os.path.isfile(fn):
            readfortrancode(fn,dowithline=dowithline,istop=0)
        else:
            outmess('readfortrancode: %s is not an existing file. Ignoring.\n'%(`fn`))
    else:
        dowithline(finalline)
    filepositiontext=''
    fin.close()
    if istop: dowithline('',1)
    else:
        gotnextfile,filepositiontext,currentfilename,sourcecodeform,strictf77,\
           beginpattern,quite,verbose,dolowercase=saveglobals
########### Crack line
beforethisafter=r'\s*(?P<before>%s(?=\s*(\b(%s)\b)))'+ \
                          r'\s*(?P<this>(\b(%s)\b))'+ \
                          r'\s*(?P<after>%s)\s*\Z'
##
fortrantypes='character|logical|integer|real|complex|double\s*(precision\s*(complex|)|complex)|type(?=\s*\([\w\s,=(*)]*\))|byte'
typespattern=re.compile(beforethisafter%('',fortrantypes,fortrantypes,'.*')),'type'
typespattern4implicit=re.compile(beforethisafter%('',fortrantypes+'|static|automatic|undefined',fortrantypes+'|static|automatic|undefined','.*'))
#
functionpattern=re.compile(beforethisafter%('([a-z]+[\w\s(=*+-/)]*?|)','function','function','.*')),'begin'
subroutinepattern=re.compile(beforethisafter%('[a-z\s]*?','subroutine','subroutine','.*')),'begin'
#
groupbegins77=r'program|block\s*data'
beginpattern77=re.compile(beforethisafter%('',groupbegins77,groupbegins77,'.*')),'begin'
groupbegins90=groupbegins77+r'|module|interface|type(?!\s*\()'
beginpattern90=re.compile(beforethisafter%('',groupbegins90,groupbegins90,'.*')),'begin'
groupends=r'end|endprogram|endblockdata|endmodule|endinterface'
endpattern=re.compile(beforethisafter%('',groupends,groupends,'[\w\s]*')),'end'
#endifs='end\s*(if|do|where|select|while|forall)'
endifs='(end\s*(if|do|where|select|while|forall))|(module\s*procedure)'
endifpattern=re.compile(beforethisafter%('[\w]*?',endifs,endifs,'[\w\s]*')),'endif'
#
implicitpattern=re.compile(beforethisafter%('','implicit','implicit','.*')),'implicit'
dimensionpattern=re.compile(beforethisafter%('','dimension|virtual','dimension|virtual','.*')),'dimension'
externalpattern=re.compile(beforethisafter%('','external','external','.*')),'external'
optionalpattern=re.compile(beforethisafter%('','optional','optional','.*')),'optional'
requiredpattern=re.compile(beforethisafter%('','required','required','.*')),'required'
publicpattern=re.compile(beforethisafter%('','public','public','.*')),'public'
privatepattern=re.compile(beforethisafter%('','private','private','.*')),'private'
intrisicpattern=re.compile(beforethisafter%('','intrisic','intrisic','.*')),'intrisic'
intentpattern=re.compile(beforethisafter%('','intent|depend|note|check','intent|depend|note|check','\s*\(.*?\).*')),'intent'
parameterpattern=re.compile(beforethisafter%('','parameter','parameter','\s*\(.*')),'parameter'
datapattern=re.compile(beforethisafter%('','data','data','.*')),'data'
callpattern=re.compile(beforethisafter%('','call','call','.*')),'call'
callfunpattern=re.compile(beforethisafter%('','callfun','callfun','.*')),'callfun'
commonpattern=re.compile(beforethisafter%('','common','common','.*')),'common'
usepattern=re.compile(beforethisafter%('','use','use','.*')),'use'
containspattern=re.compile(beforethisafter%('','contains','contains','')),'contains'
##
groupcounter=0
grouplist={groupcounter:[]}
neededmodule=-1
expectbegin=1
skipblocksuntil=-1
def crackline(line,reset=0):
    """
    reset=-1  --- initialize
    reset=0   --- crack the line
    reset=1   --- final check if mismatch of blocks occured

    Cracked data is saved in grouplist[0].
    """
    global beginpattern,groupcounter,groupname,groupcache,grouplist,gotnextfile,\
           filepositiontext,currentfilename,neededmodule,expectbegin,skipblocksuntil,\
           skipemptyends
    if reset<0:
        groupcounter=0
        groupname={groupcounter:''}
        groupcache={groupcounter:{}}
        grouplist={groupcounter:[]}
        groupcache[groupcounter]['body']=[]
        groupcache[groupcounter]['vars']={}
        groupcache[groupcounter]['block']=''
        groupcache[groupcounter]['name']=''
        neededmodule=-1
        skipblocksuntil=-1
        return
    if reset>0:
        fl=0
        if f77modulename and neededmodule==groupcounter: fl=2
        while groupcounter>fl:
            outmess('crackline: groupcounter=%s groupname=%s\n'%(`groupcounter`,`groupname`))
            outmess('crackline: Mismatch of blocks encountered. Trying to fix it by assuming "end" statement.\n')
            grouplist[groupcounter-1].append(groupcache[groupcounter])
            grouplist[groupcounter-1][-1]['body']=grouplist[groupcounter]
            del grouplist[groupcounter]
            groupcounter=groupcounter-1
        if f77modulename and neededmodule==groupcounter:
            grouplist[groupcounter-1].append(groupcache[groupcounter])
            grouplist[groupcounter-1][-1]['body']=grouplist[groupcounter]
            del grouplist[groupcounter]
            groupcounter=groupcounter-1 # end interface
            grouplist[groupcounter-1].append(groupcache[groupcounter])
            grouplist[groupcounter-1][-1]['body']=grouplist[groupcounter]
            del grouplist[groupcounter]
            groupcounter=groupcounter-1 # end module
            neededmodule=-1
        return
    if line=='': return
    flag=0
    for pat in [dimensionpattern,externalpattern,intentpattern,optionalpattern,
                requiredpattern,
                parameterpattern,datapattern,publicpattern,privatepattern,
                intrisicpattern,
                endifpattern,endpattern,
                beginpattern,functionpattern,subroutinepattern,
                implicitpattern,typespattern,commonpattern,
                callpattern,usepattern,containspattern
                ]:
        m = pat[0].match(line)
        if m: break
        flag=flag+1
    if not m:
        if 0<=skipblocksuntil<=groupcounter:return
        if groupcache[groupcounter].has_key('externals'):
            for name in groupcache[groupcounter]['externals']:
                if invbadnames.has_key(name):
                    name=invbadnames[name]
                if groupcache[groupcounter].has_key('interfaced') and name in groupcache[groupcounter]['interfaced']: continue
                m1=re.match(r'(?P<before>[^"]*)\b%s\b\s*@\(@(?P<args>[^@]*)@\)@.*\Z'%name,markouterparen(line))
                if m1:
                    m2=re.match(r'\s*(?P<result>\b[a-z]+[\w]*\b)\s*[=].*',m1.group('before'))
                    if m2:
                        line='callfun %s(%s) %s'%(name,m1.group('args'),m2.group('result'))
                    else: line='callfun %s(%s)'%(name,m1.group('args'))
                    m = callfunpattern[0].match(line)
                    if not m:
                        outmess('crackline: could not resolve function call for line=%s.\n'%`line`)
                        return
                    analyzeline(m,'callfun',line)
                    return
        if verbose:
            outmess('crackline:%d: No pattern for line\n'%(groupcounter))
        return
    elif pat[1]=='end':
        if 0<=skipblocksuntil<groupcounter:
            groupcounter=groupcounter-1
            if skipblocksuntil<=groupcounter: return
        if groupcounter<=0:
            raise 'crackline: groupcounter is nonpositive. Check the blocks.'
        m1 = beginpattern[0].match((line))
        if (m1) and (not m1.group('this')==groupname[groupcounter]):
            raise 'crackline: End group %s does not match with previous Begin group %s\n\t%s'%(`m1.group('this')`,`groupname[groupcounter]`,filepositiontext)
        if skipblocksuntil==groupcounter:
            skipblocksuntil=-1
        grouplist[groupcounter-1].append(groupcache[groupcounter])
        grouplist[groupcounter-1][-1]['body']=grouplist[groupcounter]
        del grouplist[groupcounter]
        groupcounter=groupcounter-1
        if not skipemptyends:
            expectbegin=1
    elif pat[1] == 'begin':
        if 0<=skipblocksuntil<=groupcounter:
            groupcounter=groupcounter+1
            return
        gotnextfile=0
        analyzeline(m,pat[1],line)
        expectbegin=0
    elif pat[1]=='endif':
        pass
    elif pat[1]=='contains':
        if ignorecontains: return
        if 0<=skipblocksuntil<=groupcounter: return
        skipblocksuntil=groupcounter
    else:
        if 0<=skipblocksuntil<=groupcounter:return
        analyzeline(m,pat[1],line)
def markouterparen(line):
    l='';f=0
    for c in line:
        if c=='(':
            f=f+1
            if f==1: l=l+'@(@'; continue
        elif c==')':
            f=f-1
            if f==0: l=l+'@)@'; continue
        l=l+c
    return l
def markoutercomma(line,comma=','):
    l='';f=0
    for c in line:
        if c=='(':f=f+1
        elif c==')':f=f-1
        elif c==comma and f==0: l=l+'@'+comma+'@'; continue
        l=l+c
    return l
def unmarkouterparen(line):
    return string.replace(string.replace(line,'@(@','('),'@)@',')')
def appenddecl(decl,decl2,force=1):
    if not decl: decl={}
    if not decl2: return decl
    if decl is decl2: return decl
    for k in decl2.keys():
        if k=='typespec':
            if force or not decl.has_key(k): decl[k]=decl2[k]
        elif k=='attrspec':
            for l in decl2[k]:
                decl=setattrspec(decl,l,force)
        elif k=='kindselector':
            decl=setkindselector(decl,decl2[k],force)
        elif k=='charselector':
            decl=setcharselector(decl,decl2[k],force)
        elif k in ['=','typename']:
            if force or not decl.has_key(k): decl[k]=decl2[k]
        elif k=='note':
            pass
        elif k in ['intent','check','dimension','optional','required']:
            errmess('appenddecl: "%s" not implemented.\n'%k)
        else:
            raise 'appenddecl: Unknown variable definition key:', k
    return decl

selectpattern=re.compile(r'\s*(?P<this>(@\(@.*?@\)@|[*][\d*:]+|[*]\s*@\(@.*?@\)@|))(?P<after>.*)\Z')
nameargspattern=re.compile(r'\s*(?P<name>\b[\w$]+\b)\s*(\(\s*(?P<args>[\w\s,]*)\s*\)|)\s*(result(\s*\(\s*(?P<result>\b[\w$]+\b)\s*\)|))*\s*\Z')
callnameargspattern=re.compile(r'\s*(?P<name>\b[\w$]+\b)\s*(\(\s*(?P<args>[\w\s,+-/*()]*)\s*\)|)\s*(?P<result>.*)\Z')
def analyzeline(m,case,line):
    global groupcounter,groupname,groupcache,grouplist,filepositiontext,\
           currentfilename,f77modulename,neededinterface,neededmodule,expectbegin,\
           gotnextfile
    block=m.group('this')
    if expectbegin and case not in ['begin','call','callfun'] and not skipemptyends:
        newname=string.split(os.path.basename(currentfilename),'.')[0]
        outmess('analyzeline: no group yet. Creating program group with name "%s".\n'%newname)
        gotnextfile=0
        groupcounter=groupcounter+1
        groupname[groupcounter]='program'
        groupcache[groupcounter]={}
        grouplist[groupcounter]=[]
        groupcache[groupcounter]['body']=[]
        groupcache[groupcounter]['vars']={}
        groupcache[groupcounter]['block']='program'
        groupcache[groupcounter]['name']=newname
        groupcache[groupcounter]['from']='fromsky'
        expectbegin=0
    if case in ['begin','call','callfun']:
        # Crack line => block,name,args,result
        if re.match(r'block\s*data',block): block='block data'
        if case in ['call','callfun']: m1=callnameargspattern.match(m.group('after'))
        else: m1=nameargspattern.match(m.group('after'))
        if not m1:
            if block not in ['interface','block data']:                
                outmess('analyzeline: No name/args pattern found for line.\n')
            if block=='block data':
                name='_BLOCK_DATA_'
            else:
                name=''
            result=None
            args=[]
        else:
            name=rmbadname1(m1.group('name'))
            result=m1.group('result')
            args=m1.group('args')
        if args: args=rmbadname(map(string.strip,string.split(markoutercomma(args),'@,@')))
        else: args=[]
        # end of crack line => block,name,args,result
        needmodule=0
        needinterface=0
        if case in ['call','callfun']:
            needinterface=1
            if not groupcache[groupcounter].has_key('args'): return
            if name not in groupcache[groupcounter]['args']: return
            for it in grouplist[groupcounter]:
                if it['name']==name: return
            if name in groupcache[groupcounter]['interfaced']: return
            block={'call':'subroutine','callfun':'function'}[case]
        if f77modulename and neededmodule==-1 and groupcounter<=1:
            neededmodule=groupcounter+2
            needmodule=1
            needinterface=1        
        # Create new block(s)
        groupcounter=groupcounter+1
        groupcache[groupcounter]={}
        grouplist[groupcounter]=[]        
        if needmodule:
            if verbose:
                outmess('analyzeline: Creating module block %s\n'%`f77modulename`,0)
            groupname[groupcounter]='module'
            groupcache[groupcounter]['block']='module'
            groupcache[groupcounter]['name']=f77modulename
            groupcache[groupcounter]['from']=''
            groupcache[groupcounter]['body']=[]
            groupcache[groupcounter]['externals']=[]
            groupcache[groupcounter]['interfaced']=[]
            groupcache[groupcounter]['vars']={}
            groupcounter=groupcounter+1
            groupcache[groupcounter]={}
            grouplist[groupcounter]=[]
        if needinterface:
            if verbose:
                outmess('analyzeline: Creating additional interface block.\n',0)
            groupname[groupcounter]='interface'
            groupcache[groupcounter]['block']='interface'
            groupcache[groupcounter]['name']='unknown_interface'
            groupcache[groupcounter]['from']='%s:%s'%(groupcache[groupcounter-1]['from'],groupcache[groupcounter-1]['name'])
            groupcache[groupcounter]['body']=[]
            groupcache[groupcounter]['externals']=[]
            groupcache[groupcounter]['interfaced']=[]
            groupcache[groupcounter]['vars']={}
            groupcounter=groupcounter+1
            groupcache[groupcounter]={}
            grouplist[groupcounter]=[]
        groupname[groupcounter]=block
        groupcache[groupcounter]['block']=block
        if not name: name='unknown_'+block
        groupcache[groupcounter]['prefix']=m.group('before')
        groupcache[groupcounter]['name']=rmbadname1(name)
        groupcache[groupcounter]['result']=result
        if groupcounter==1:
            groupcache[groupcounter]['from']=currentfilename
        else:
            if f77modulename and groupcounter==3:
                groupcache[groupcounter]['from']='%s:%s'%(groupcache[groupcounter-1]['from'],currentfilename)
            else:
                groupcache[groupcounter]['from']='%s:%s'%(groupcache[groupcounter-1]['from'],groupcache[groupcounter-1]['name'])
        for k in groupcache[groupcounter].keys():
            if not groupcache[groupcounter][k]: del groupcache[groupcounter][k]
        groupcache[groupcounter]['args']=args
        groupcache[groupcounter]['body']=[]
        groupcache[groupcounter]['externals']=[]
        groupcache[groupcounter]['interfaced']=[]
        groupcache[groupcounter]['vars']={}
        # end of creation
        if case in ['call','callfun']: # set parents variables
            if name not in groupcache[groupcounter-2]['externals']:
                groupcache[groupcounter-2]['externals'].append(name)
            groupcache[groupcounter]['vars']=copy.deepcopy(groupcache[groupcounter-2]['vars'])
            #try: del groupcache[groupcounter]['vars'][groupcache[groupcounter-2]['name']]
            #except: pass
            try: del groupcache[groupcounter]['vars'][name][groupcache[groupcounter]['vars'][name]['attrspec'].index('external')]
            except: pass
        if block in ['function','subroutine']: # set global attributes
            try: groupcache[groupcounter]['vars'][name]=appenddecl(groupcache[groupcounter]['vars'][name],groupcache[groupcounter-2]['vars'][''])
            except: pass
            if case=='callfun': # return type
                if result and groupcache[groupcounter]['vars'].has_key(result):
                    if not name==result:
                        groupcache[groupcounter]['vars'][name]=appenddecl(groupcache[groupcounter]['vars'][name],groupcache[groupcounter]['vars'][result])
            #if groupcounter>1: # name is interfaced
            try: groupcache[groupcounter-2]['interfaced'].append(name)
            except: pass
        if block=='function':
            t=typespattern[0].match(m.group('before')+' '+name)
            if t:
                typespec,selector,attr,edecl=cracktypespec0(t.group('this'),t.group('after'))
                updatevars(typespec,selector,attr,edecl)
        if case in ['call','callfun']:
            grouplist[groupcounter-1].append(groupcache[groupcounter])
            grouplist[groupcounter-1][-1]['body']=grouplist[groupcounter]
            del grouplist[groupcounter]
            groupcounter=groupcounter-1 # end routine
            grouplist[groupcounter-1].append(groupcache[groupcounter])
            grouplist[groupcounter-1][-1]['body']=grouplist[groupcounter]
            del grouplist[groupcounter]
            groupcounter=groupcounter-1 # end interface
    elif case=='type':
        typespec,selector,attr,edecl=cracktypespec0(block,m.group('after'))
        updatevars(typespec,selector,attr,edecl)
    elif case in ['dimension','intent','optional','required','external','public','private','intrisic','check']:
        edecl=groupcache[groupcounter]['vars']
        ll=m.group('after')
        i=string.find(ll,'::')
        if i<0 and case=='intent':
            i=string.find(markouterparen(ll),'@)@')-2
            ll=ll[:i+1]+'::'+ll[i+1:]
            i=string.find(ll,'::')
        if i<0:i=0;pl=''
        else: pl=string.strip(ll[:i]);ll=ll[i+2:]
        for e in map(string.strip,string.split(markoutercomma(ll),'@,@')):
            m1=namepattern.match(e)
            if not m1:
                if case in ['public','private']: k=''
                else:
                    print m.groupdict()
                    outmess('analyzeline: no name pattern found in %s statement for %s. Skipping.\n'%(case,`e`));continue
            else:
                k=rmbadname1(m1.group('name'))
            if not edecl.has_key(k): edecl[k]={}
            if case=='dimension': ap=case+m1.group('after')
            if case=='intent': ap=m.group('this')+pl
            if case in ['optional','required','public','external','private','intrisic']: ap=case
            if edecl[k].has_key('attrspec'): edecl[k]['attrspec'].append(ap)
            else: edecl[k]['attrspec']=[ap]
            if case=='external':
                if not groupcache[groupcounter].has_key('externals'):
                    groupcache[groupcounter]['externals']=[]
                groupcache[groupcounter]['externals'].append(k)
        groupcache[groupcounter]['vars']=edecl
    elif case=='parameter':
        edecl=groupcache[groupcounter]['vars']
        ll=string.strip(m.group('after'))[1:-1]
        for e in string.split(markoutercomma(ll),'@,@'):
            try: k,initexpr=map(string.strip,string.split(e,'='))
            except:
                outmess('analyzeline: could not extract name,expr in parameter statement "%s" of "%s"\n'%(e,ll));continue
            k=rmbadname1(k)
            if not edecl.has_key(k): edecl[k]={}
            if edecl[k].has_key('=') and (not edecl[k]['=']==initexpr):
                outmess('analyzeline: Overwriting the value of parameter "%s" ("%s") with "%s".\n'%(k,edecl[k]['='],initexpr))
            edecl[k]['=']=initexpr
            if edecl[k].has_key('attrspec'): edecl[k]['attrspec'].append('parameter')
            else: edecl[k]['attrspec']=['parameter']
        groupcache[groupcounter]['vars']=edecl
    elif case=='implicit':
        if string.strip(m.group('after'))=='none':
            groupcache[groupcounter]['implicit']=None
        elif m.group('after'):
            if groupcache[groupcounter].has_key('implicit'):
                impl=groupcache[groupcounter]['implicit']
            else: impl={}
            if impl is None:
                outmess('analyzeline: Overwriting earlier "implicit none" statement.\n')
                impl={}
            for e in string.split(markoutercomma(m.group('after')),'@,@'):
                decl={}
                m1=re.match(r'\s*(?P<this>.*?)\s*(\(\s*(?P<after>[a-z-, ]+)\s*\)\s*|)\Z',e)
                if not m1:
                    outmess('analyzeline: could not extract info of implicit statement part "%s"\n'%(e));continue
                m2=typespattern4implicit.match(m1.group('this'))
                if not m2:
                    outmess('analyzeline: could not extract types pattern of implicit statement part "%s"\n'%(e));continue
                typespec,selector,attr,edecl=cracktypespec0(m2.group('this'),m2.group('after'))
                kindselect,charselect,typename=cracktypespec(typespec,selector)
                decl['typespec']=typespec
                decl['kindselector']=kindselect
                decl['charselector']=charselect
                decl['typename']=typename
                for k in decl.keys():
                    if not decl[k]: del decl[k]
                for r in string.split(markoutercomma(m1.group('after')),'@,@'):
                    if '-' in r:
                        try: begc,endc=map(string.strip,string.split(r,'-'))
                        except:
                            outmess('analyzeline: expected "<char>-<char>" instead of "%s" in range list of implicit statement\n'%r);continue
                    else: begc=endc=string.strip(r)
                    if not len(begc)==len(endc)==1:
                        outmess('analyzeline: expected "<char>-<char>" instead of "%s" in range list of implicit statement (2)\n'%r);continue
                    for o in range(ord(begc),ord(endc)+1):
                        impl[chr(o)]=decl
            groupcache[groupcounter]['implicit']=impl
    elif case=='data':
        ll=[]
        dl='';il='';f=0;fc=1
        for c in m.group('after'):
            if c=="'": fc=not fc
            if c=='/' and fc: f=f+1;continue
            if f==0: dl=dl+c
            elif f==1: il=il+c
            elif f==2:
                ll.append([dl,il])
                dl=c;il='';f=0
        if f==2: ll.append([dl,il])
        vars={}
        if groupcache[groupcounter].has_key('vars'):
            vars=groupcache[groupcounter]['vars']
        for l in ll:
            l=map(string.strip,l)
            if l[0][0]==',':l[0]=l[0][1:]
            if l[0][0]=='(':
                outmess('analyzeline: implied-DO list "%s" is not supported. Skipping.\n'%l[0])
                continue
            i=0;j=0;llen=len(l[1])
            for v in rmbadname(map(string.strip,string.split(markoutercomma(l[0]),'@,@'))):
                fc=0
                while (i<llen) and (fc or not l[1][i]==','):
                    if l[1][i]=="'": fc=not fc
                    i=i+1
                i=i+1
                #v,l[1][j:i-1]=name,initvalue
                if not vars.has_key(v):
                    vars[v]={}
                if vars[v].has_key('=') and not vars[v]['=']==l[1][j:i-1]:
                    outmess('analyzeline: changing init expression of "%s" ("%s") to "%s"\n'%(v,vars[v]['='],l[1][j:i-1]))
                vars[v]['=']=l[1][j:i-1]
                j=i
        groupcache[groupcounter]['vars']=vars
    elif case=='common':
        line=string.strip(m.group('after'))
        if not line[0]=='/':line='//'+line
        cl=[]
        f=0;bn='';ol=''
        for c in line:
            if c=='/':f=f+1;continue
            if f>=3:
                if not string.strip(bn): bn='_BLNK_'
                cl.append([bn,ol])
                f=f-2;bn='';ol=''
            if f%2: bn=bn+c
            else: ol=ol+c
        if not string.strip(bn): bn='_BLNK_'
        cl.append([bn,ol])
        commonkey={}
        if groupcache[groupcounter].has_key('common'):
            commonkey=groupcache[groupcounter]['common']
        for c in cl:
            if commonkey.has_key(c[0]):
                outmess('analyzeline: previously defined common block encountered. Skipping.\n')
                continue
            commonkey[c[0]]=[]
            for i in map(string.strip,string.split(markoutercomma(c[1]),'@,@')):
                if i: commonkey[c[0]].append(i)
        groupcache[groupcounter]['common']=commonkey
    elif case=='use':
        m1=re.match(r'\A\s*(?P<name>\b[\w]+\b)\s*((,(\s*\bonly\b\s*:|(?P<notonly>))\s*(?P<list>.*))|)\s*\Z',m.group('after'))
        if m1:
            mm=m1.groupdict()
            if not groupcache[groupcounter].has_key('use'): groupcache[groupcounter]['use']={}
            name=m1.group('name')
            groupcache[groupcounter]['use'][name]={}
            isonly=0
            if mm.has_key('list') and mm['list'] is not None:
                if mm.has_key('notonly') and mm['notonly'] is None:isonly=1
                groupcache[groupcounter]['use'][name]['only']=isonly
                ll=map(string.strip,string.split(mm['list'],','))
                rl={}
                for l in ll:
                    if '=' in l:
                        m2=re.match(r'\A\s*(?P<local>\b[\w]+\b)\s*=\s*>\s*(?P<use>\b[\w]+\b)\s*\Z',l)
                        if m2: rl[string.strip(m2.group('local'))]=string.strip(m2.group('use'))
                        else:
                            outmess('analyzeline: Not local=>use pattern found in %s\n'%`l`)
                    else:
                        rl[l]=l
                    groupcache[groupcounter]['use'][name]['map']=rl
            else:
                pass
            
        else:
            print m.groupdict()
            outmess('analyzeline: Could not crack the use statement.\n')
    else:
        if verbose:
            print m.groupdict()
            outmess('analyzeline: No code implemented for line.\n')
def cracktypespec0(typespec,ll):
    selector=None
    attr=None
    if re.match(r'double\s*complex',typespec): typespec='double complex'
    elif re.match(r'double\s*precision',typespec): typespec='double precision'
    else: typespec=string.strip(typespec)
    m1=selectpattern.match(markouterparen(ll))
    if not m1:
        outmess('cracktypespec0: no kind/char_selector pattern found for line.\n')
        return
    d=m1.groupdict()
    for k in d.keys(): d[k]=unmarkouterparen(d[k])
    if typespec in ['complex','integer','logical','real','character','type']:
        selector=d['this']
        ll=d['after']
    i=string.find(ll,'::')
    if i>=0:
        attr=string.strip(ll[:i])
        ll=ll[i+2:]
    return typespec,selector,attr,ll
#####
namepattern=re.compile(r'\s*(?P<name>\b[\w]+\b)\s*(?P<after>.*)\s*\Z')
kindselector=re.compile(r'\s*(\(\s*(kind\s*=)?\s*(?P<kind>.*)\s*\)|[*]\s*(?P<kind2>.*?))\s*\Z')
charselector=re.compile(r'\s*(\((?P<lenkind>.*)\)|[*]\s*(?P<charlen>.*))\s*\Z')
lenkindpattern=re.compile(r'\s*(kind\s*=\s*(?P<kind>.*?)\s*(@,@\s*len\s*=\s*(?P<len>.*)|)|(len\s*=\s*|)(?P<len2>.*?)\s*(@,@\s*(kind\s*=\s*|)(?P<kind2>.*)|))\s*\Z')
lenarraypattern=re.compile(r'\s*(@\(@\s*(?!/)\s*(?P<array>.*?)\s*@\)@\s*[*]\s*(?P<len>.*?)|([*]\s*(?P<len2>.*?)|)\s*(@\(@\s*(?!/)\s*(?P<array2>.*?)\s*@\)@|))\s*(=\s*(?P<init>.*?)|(@\(@|)/\s*(?P<init2>.*?)\s*/(@\)@|)|)\s*\Z')
def removespaces(expr):
    expr=string.strip(expr)
    if len(expr)<=1: return expr
    expr2=expr[0]
    for i in range(1,len(expr)-1):
        if expr[i]==' ' and \
           ((expr[i+1] in "()[]{}= ") or (expr[i-1] in "()[]{}= ")): continue
        expr2=expr2+expr[i]
    expr2=expr2+expr[-1]
    return expr2
def markinnerspaces(line):
    l='';f=0
    cc='\''
    cc1='"'
    cb=''
    for c in line:
        if cb=='\\' and c in ['\\','\'','"']:
            l=l+c;
            cb=c
            continue 
        if f==0 and c in ['\'','"']: cc=c; cc1={'\'':'"','"':'\''}[c]
        if c==cc:f=f+1
        elif c==cc:f=f-1
        elif c==' ' and f==1: l=l+'@_@'; continue
        l=l+c;cb=c
    return l
def updatevars(typespec,selector,attrspec,entitydecl):
    global groupcache,groupcounter
    kindselect,charselect,typename=cracktypespec(typespec,selector)
    if attrspec:
        attrspec=map(string.strip,string.split(markoutercomma(attrspec),'@,@'))
    el=map(string.strip,string.split(markoutercomma(entitydecl),'@,@'))
    el1=[]
    for e in el:
        for e1 in map(string.strip,string.split(markoutercomma(removespaces(markinnerspaces(e)),comma=' '),'@ @')):
            if e1: el1.append(string.replace(e1,'@_@',' '))
    for e in el1:
        m=namepattern.match(e)
        if not m:
            outmess('updatevars: no name pattern found for entity=%s. Skipping.\n'%(`e`))
            continue
        ename=rmbadname1(m.group('name'))
        edecl={}
        if groupcache[groupcounter]['vars'].has_key(ename):
            edecl=groupcache[groupcounter]['vars'][ename]
            if not edecl.has_key('typespec'):
                edecl['typespec']=typespec
            elif typespec and (not typespec==edecl['typespec']):
                outmess('updatevars: attempt to change the type of "%s" ("%s") to "%s". Ignoring.\n' % (ename,edecl['typespec'],typespec))
            if not edecl.has_key('kindselector'):
                edecl['kindselector']=copy.copy(kindselect)
            elif kindselect:
                for k in kindselect.keys():
                    if edecl['kindselector'].has_key(k) and (not kindselect[k]==edecl['kindselector'][k]):
                        outmess('updatevars: attempt to change the kindselector "%s" of "%s" ("%s") to "%s". Ignoring.\n' % (k,ename,edecl['kindselector'][k],kindselect[k]))
                    else: edecl['kindselector'][k]=copy.copy(kindselect[k])
            if not edecl.has_key('charselector'):
                edecl['charselector']=charselect
            elif charselect:
                for k in charselect.keys():
                    if edecl['charselector'].has_key(k) and (not charselect[k]==edecl['charselector'][k]):
                        outmess('updatevars: attempt to change the charselector "%s" of "%s" ("%s") to "%s". Ignoring.\n' % (k,ename,edecl['charselector'][k],charselect[k]))
                    else: edecl['charselector'][k]=copy.copy(charselect[k])
            if not edecl.has_key('typename'):
                edecl['typename']=typename
            elif typename and (not edecl['typename']==typename):
                outmess('updatevars: attempt to change the typename of "%s" ("%s") to "%s". Ignoring.\n' % (ename,edecl['typename'],typename))
            if not edecl.has_key('attrspec'):
                edecl['attrspec']=copy.copy(attrspec)
            elif attrspec:
                for a in attrspec:
                    if a not in edecl['attrspec']:
                        edecl['attrspec'].append(a)
        else:
            edecl['typespec']=copy.copy(typespec)
            edecl['kindselector']=copy.copy(kindselect)
            edecl['charselector']=copy.copy(charselect)
            edecl['typename']=typename
            edecl['attrspec']=copy.copy(attrspec)
        if m.group('after'):
            m1=lenarraypattern.match(markouterparen(m.group('after')))
            if m1:
                d1=m1.groupdict()
                for lk in ['len','array','init']:
                    if d1[lk+'2'] is not None: d1[lk]=d1[lk+'2']; del d1[lk+'2']
                for k in d1.keys():
                    if d1[k] is not None: d1[k]=unmarkouterparen(d1[k])
                    else: del d1[k]
                if d1.has_key('len') and d1.has_key('array'):
                    if d1['len']=='':
                        d1['len']=d1['array']
                        del d1['array']
                    else:
                        d1['array']=d1['array']+','+d1['len']
                        del d1['len']
                        errmess('updatevars: "%s %s" is mapped to "%s %s(%s)"\n'%(typespec,e,typespec,ename,d1['array']))
                if d1.has_key('array'):
                    if not edecl.has_key('attrspec') or (not edecl['attrspec']):
                        edecl['attrspec']=['dimension(%s)'%d1['array']]
                    else:
                        edecl['attrspec'].append('dimension(%s)'%d1['array'])
                if d1.has_key('len'):
                    if typespec in ['complex','integer','logical','real']:
                        if (not edecl.has_key('kindselector')) or (not edecl['kindselector']):
                            edecl['kindselector']={}
                        edecl['kindselector']['*']=d1['len']
                    elif typespec == 'character':
                        if (not edecl.has_key('charselector')) or (not edecl['charselector']): edecl['charselector']={}
                        if edecl['charselector'].has_key('len'): del edecl['charselector']['len']
                        edecl['charselector']['*']=d1['len']
                if d1.has_key('init'):
                    if edecl.has_key('=') and (not edecl['=']==d1['init']):
                        outmess('updatevars: attempt to change the init expression of "%s" ("%s") to "%s". Ignoring.\n' % (ename,edecl['='],d1['init']))
                    else:
                        edecl['=']=d1['init']
            else:
                outmess('updatevars: could not crack entity declaration "%s". Ignoring.\n'%(ename+m.group('after')))
        for k in edecl.keys():
            if not edecl[k]: del edecl[k]
        groupcache[groupcounter]['vars'][ename]=edecl
def cracktypespec(typespec,selector):
    kindselect=None
    charselect=None
    typename=None
    if selector:
        if typespec in ['complex','integer','logical','real']:
            kindselect=kindselector.match(selector)
            if not kindselect:
                outmess('cracktypespec: no kindselector pattern found for %s\n'%(`selector`))
                return
            kindselect=kindselect.groupdict()
            kindselect['*']=kindselect['kind2']
            del kindselect['kind2']
            for k in kindselect.keys():
                if not kindselect[k]: del kindselect[k]
        elif typespec=='character':
            charselect=charselector.match(selector)
            if not charselect:
                outmess('cracktypespec: no charselector pattern found for %s\n'%(`selector`))
                return
            charselect=charselect.groupdict()
            charselect['*']=charselect['charlen']
            del charselect['charlen']
            if charselect['lenkind']:
                lenkind=lenkindpattern.match(markoutercomma(charselect['lenkind']))
                lenkind=lenkind.groupdict()
                for lk in ['len','kind']:
                    if lenkind[lk+'2']:
                        lenkind[lk]=lenkind[lk+'2']
                    charselect[lk]=lenkind[lk]
                    del lenkind[lk+'2']
            del charselect['lenkind']
            for k in charselect.keys():
                if not charselect[k]: del charselect[k]
        elif typespec=='type':
            typename=re.match(r'\s*\(\s*(?P<name>\w+)\s*\)',selector)
            if typename: typename=typename.group('name')
            else: outmess('cracktypespec: no typename found in %s\n'%(`typespec+selector`))
        else:
            outmess('cracktypespec: no selector used for %s\n'%(`selector`))
    return kindselect,charselect,typename
######
def setattrspec(decl,attr,force=0):
    if not decl: decl={}
    if not attr: return decl
    if not decl.has_key('attrspec'):
        decl['attrspec']=[attr]
        return decl
    if force: decl['attrspec'].append(attr)
    if attr in decl['attrspec']: return decl
    if attr=='static' and 'automatic' not in decl['attrspec']:
        decl['attrspec'].append(attr)
    elif attr=='automatic' and 'static' not in decl['attrspec']:
        decl['attrspec'].append(attr)
    elif attr=='public' and 'private' not in decl['attrspec']:
        decl['attrspec'].append(attr)
    elif attr=='private' and 'public' not in decl['attrspec']:
        decl['attrspec'].append(attr)
    else:
        decl['attrspec'].append(attr)
    return decl
def setkindselector(decl,sel,force=0):
    if not decl: decl={}
    if not sel: return decl
    if not decl.has_key('kindselector'):
        decl['kindselector']=sel
        return decl
    for k in sel.keys():
        if force or not decl['kindselector'].has_key(k):
            decl['kindselector'][k]=sel[k]
    return decl
def setcharselector(decl,sel,force=0):
    if not decl: decl={}
    if not sel: return decl
    if not decl.has_key('charselector'):
        decl['charselector']=sel
        return decl
    for k in sel.keys():
        if force or not decl['charselector'].has_key(k):
            decl['charselector'][k]=sel[k]
    return decl
def getblockname(block,unknown='unknown'):
    if block.has_key('name'): return block['name']
    return unknown
###### post processing
def setmesstext(block):
    global filepositiontext
    try: filepositiontext='In: %s:%s\n'%(block['from'],block['name'])
    except: pass
usermodules=[]
def postcrack(block,args=None,tab=''):
    """
    TODO:
          function return values
          common blocks
          determine expression types if in argument list
    """
    global usermodules,onlyfunctions
    if type(block) is type([]):
        gret=[]
        uret=[]
        for g in block:
            setmesstext(g)
            g=postcrack(g,tab=tab+'\t')
            if g.has_key('name') and string.find(g['name'],'__user__')>=0: # sort user routines to appear first
                uret.append(g)
            else:
                gret.append(g)
        return uret+gret
    setmesstext(block)
    if not type(block) is type({}) and not block.has_key('block'):
        raise 'postcrack: Expected block dictionary instead of ',block
    if block.has_key('name') and not block['name']=='unknown_interface':
        outmess('%sBlock: %s\n'%(tab,block['name']),0)
    blocktype=block['block']
    block=analyzeargs(block)
    block=analyzecommon(block)
    block['vars']=analyzevars(block)
    block['sortvars']=sortvarnames(block['vars'])
    if block.has_key('args') and block['args']:
        args=block['args']
    block['body']=analyzebody(block,args,tab=tab)
    userisdefined=[]
    if block.has_key('use'):
        useblock=block['use']
        for k in useblock.keys():
            if string.find(k,'__user__')>=0:
                userisdefined.append(k)
    else: useblock={}
    name=''
    if block.has_key('name'):name=block['name']
    if block.has_key('externals') and block['externals']:# and not userisdefined: # Build a __user__ module
        interfaced=[]
        if block.has_key('interfaced'): interfaced=block['interfaced']
        mvars=copy.copy(block['vars'])
        if name: mname=name+'__user__routines'
        else: mname='unknown__user__routines'
        if mname in userisdefined:
            i=1
            while '%s_%i'%(mname,i) in userisdefined: i=i+1
            mname='%s_%i'%(mname,i)
        interface={'block':'interface','body':[],'vars':{},'name':name+'_user_interface'}
        for e in block['externals']:
            if e in interfaced:
                edef=[]
                j=-1
                for b in block['body']:
                    j=j+1
                    if b['block']=='interface':
                        i=-1
                        for bb in b['body']:
                            i=i+1
                            if bb.has_key('name') and bb['name']==e:
                                edef=copy.copy(bb)
                                del b['body'][i]
                                break
                        if edef:
                            if not b['body']: del block['body'][j]
                            del interfaced[interfaced.index(e)]
                            break
                interface['body'].append(edef)
            else:
                if mvars.has_key(e):
                    interface['vars'][e]=mvars[e]
        if interface['vars'] or interface['body']:
            block['interfaced']=interfaced
            mblock={'block':'module','body':[interface],'vars':{},'name':mname,'interfaced':block['externals']}
            useblock[mname]={}
            usermodules.append(mblock)
    if useblock:
        block['use']=useblock
    return block
def sortvarnames(vars):
    indep = []
    dep = []
    for v in vars.keys():
        if vars[v].has_key('depend') and vars[v]['depend']: dep.append(v)
        else: indep.append(v)
    ns=dep
    n=len(ns)
    for i in range(n):
        for j in range(i+1,n):
            for k in range(i,j+1):
                if (vars[ns[i]].has_key('depend') and ns[k] in vars[ns[i]]['depend']):
                    c=ns[k]
                    ns[k]=ns[i]
                    ns[i]=c
    ns = indep + ns
    return ns
def analyzecommon(block):
    if not hascommon(block): return block
    commonvars=[]
    for k in block['common'].keys():
        comvars=[]
        for e in block['common'][k]:
            m=re.match(r'\A\s*\b(?P<name>.*?)\b\s*(\((?P<dims>.*?)\)|)\s*\Z',e)
            if m:
                dims=[]
                if m.group('dims'):
                    dims=map(string.strip,string.split(markoutercomma(m.group('dims')),'@,@'))
                n=string.strip(m.group('name'))
                if block['vars'].has_key(n):
                    if block['vars'][n].has_key('attrspec'):
                        block['vars'][n]['attrspec'].append('dimension(%s)'%(string.join(dims,',')))
                    else:
                        block['vars'][n]['attrspec']=['dimension(%s)'%(string.join(dims,','))]
                else:
                    if dims:
                        block['vars'][n]={'attrspec':['dimension(%s)'%(string.join(dims,','))]}
                    else: block['vars'][n]={}
                if n not in commonvars: commonvars.append(n)
            else:
                n=e
                errmess('analyzecommon: failed to extract "<name>[(<dims>)]" from "%s" in common /%s/.\n'%(e,k))
            comvars.append(n)
        block['common'][k]=comvars
    if not block.has_key('commonvars'):
        block['commonvars']=commonvars
    else:
        block['commonvars']=block['commonvars']+commonvars
    return block
def analyzebody(block,args,tab=''):
    global usermodules,skipfuncs,onlyfuncs
    setmesstext(block)
    body=[]
    for b in block['body']:
        if b['block'] in ['function','subroutine']:
            if args is not None and b['name'] not in args:
                continue
            else:
                as=b['args']
            if b['name'] in skipfuncs:
                continue
            if onlyfuncs and b['name'] not in onlyfuncs:
                continue
        else: as=args
        b=postcrack(b,as,tab=tab+'\t')
        if b['block']=='interface' and not b['body']:
            continue
        if b['block']=='module':
            usermodules.append(b)
        else:
            body.append(b)
    return body
def buildimplicitrules(block):
    setmesstext(block)
    implicitrules=defaultimplicitrules
    attrrules={}
    if block.has_key('implicit'):
        if block['implicit'] is None:
            implicitrules=None
            if verbose:
                outmess('buildimplicitrules: no implicit rules for routine %s.\n'%`block['name']`)
        else:
            for k in block['implicit'].keys():
                if block['implicit'][k]['typespec'] not in ['static','automatic']:
                    implicitrules[k]=block['implicit'][k]
                else:
                    attrrules[k]=block['implicit'][k]['typespec']
    return implicitrules,attrrules
def getlincoef(e,xset): # e = a*x+b ; x in xset
    try:
        c = int(eval(e))
        return 0,c,None
    except: pass
    for x in xset:
        m = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',e)
        if m:
            try:
                m1 = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',e)
                while m1:
                    ee = '%s(%s)%s'%(m1.group('before'),0,m1.group('after'))
                    m1 = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',ee)
                b = eval(ee)
                m1 = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',e)
                while m1:
                    ee = '%s(%s)%s'%(m1.group('before'),1,m1.group('after'))
                    m1 = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',ee)
                a = eval(ee) - b
                m1 = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',e)
                while m1:
                    ee = '%s(%s)%s'%(m1.group('before'),0.5,m1.group('after'))
                    m1 = re.match(r'(?P<before>.*?)\b'+x+r'\b(?P<after>.*)',ee)
                c = eval(ee)
                if (a*0.5+b==c):
                    return a,b,x
            except: pass
            break
    return None,None,None
def getarrlen(dl,args):
    d = '%s-(%s)+1'%(dl[1],dl[0])
    try: return `eval(d)`,None,None
    except: pass
    d1,d2=getlincoef(dl[0],args),getlincoef(dl[1],args)
    if None not in [d1[0],d2[0]]:
        if (d1[0],d2[0])==(0,0):
            return `d2[1]-d1[1]+1`,None,None
        if d1[0]==0 and d2[2] in args:
            b = d2[1] - d1[1] + 1
            if b<0: return '%s * %s - %s'%(d2[0],d2[2],-b),d2[2],'+%s)/(%s)'%(-b,d2[0])
            elif b: return '%s * %s + %s'%(d2[0],d2[2],b),d2[2],'-%s)/(%s)'%(b,d2[0])
            else: return '%s * %s'%(d2[0],d2[2]),d2[2],')/(%s)'%(d2[0])
        if d2[0]==0 and d1[2] in args:
            b = d2[1] - d1[1] + 1
            if b<0: return '%s * %s - %s'%(-d1[0],d1[2],-b),d1[2],'+%s)/(%s)'%(-b,-d1[0])
            elif b: return '%s * %s + %s'%(-d1[0],d1[2],b),d1[2],'-%s)/(%s)'%(b,-d1[0])
            else: return '%s * %s'%(-d1[0],d1[2]),d1[2],')/(%s)'%(-d1[0])
        if d1[2]==d2[2] and d1[2] in args:
            b = d2[1] - d1[1] + 1
            a = d2[0] - d1[0]
            if not a: return `b`,None,None
            if b<0: return '%s * %s - %s'%(a,d1[2],-b),d2[2],'+%s)/(%s)'%(-b,a)
            elif b: return '%s * %s + %s'%(a,d1[2],b),d2[2],'-%s)/(%s)'%(b,a)
            else: return '%s * %s'%(a,d1[2]),d2[2],')/(%s)'%(a)
    return d,None,None
def analyzevars(block):
    setmesstext(block)
    implicitrules,attrrules=buildimplicitrules(block)
    vars=copy.copy(block['vars'])
    if block['block']=='function' and not vars.has_key(block['name']):
        vars[block['name']]={}
    if block['vars'].has_key(''):
        del vars['']
        if block['vars'][''].has_key('attrspec'):
            gen=block['vars']['']['attrspec']
            for n in vars.keys():
                for k in ['public','private']:
                    if k in gen:
                        vars[n]=setattrspec(vars[n],k)
    svars=[]
    for a in block['args']:
        if vars.has_key(a): svars.append(a)
    for n in vars.keys():
        if n not in svars: svars.append(n)
    for n in svars:
        if n[0] in attrrules.keys():
            vars[n]=setattrspec(vars[n],attrrules[n[0]])
        if not vars[n].has_key('typespec'):
            if implicitrules:
                for k in implicitrules[n[0]].keys():
                    if k=='typespec' and implicitrules[n[0]][k]=='undefined':
                        continue
                    if not vars[n].has_key(k):
                        vars[n][k]=implicitrules[n[0]][k]
                    elif k=='attrspec':
                        for l in implicitrules[n[0]][k]:
                            vars[n]=setattrspec(vars[n],l)
            else:
                if not vars[n].has_key('attrspec') or 'external' not in vars[n]['attrspec']:
                    if n in block['args']:
                        outmess('analyzevars: typespec of variable %s is not defined in routine %s.\n'%(`n`,block['name']))
        savelindims = {}
        if vars[n].has_key('attrspec'):
            attr=vars[n]['attrspec']
            attr.reverse()
            vars[n]['attrspec']=[]
            dim,intent,depend,check,note=None,None,None,None,None
            for a in attr:
                if a[:9]=='dimension': dim=(string.strip(a[9:]))[1:-1]
                elif a[:6]=='intent': intent=(string.strip(a[6:]))[1:-1]
                elif a[:6]=='depend': depend=(string.strip(a[6:]))[1:-1]
                elif a[:5]=='check': check=(string.strip(a[5:]))[1:-1]
                elif a[:4]=='note': note=(string.strip(a[4:]))[1:-1]
                else: vars[n]=setattrspec(vars[n],a)
                if intent:
                    if not vars[n].has_key('intent'): vars[n]['intent']=[]
                    for c in map(string.strip,string.split(markoutercomma(intent),'@,@')):
                        if not c in vars[n]['intent']:
                            vars[n]['intent'].append(c)
                    intent=None
                if note:
                    note=string.replace(note,'\\n\\n','\n\n')
                    note=string.replace(note,'\\n ','\n')
                    if not vars[n].has_key('note'): vars[n]['note']=[note]
                    else: vars[n]['note'].append(note)
                    note=None
                if depend is not None:
                    if not vars[n].has_key('depend'): vars[n]['depend']=[]
                    for c in rmbadname(map(string.strip,string.split(markoutercomma(depend),'@,@'))):
                        if c not in vars[n]['depend']:
                            vars[n]['depend'].append(c)
                    depend=None
                if check is not None:
                    if not vars[n].has_key('check'): vars[n]['check']=[]
                    for c in map(string.strip,string.split(markoutercomma(check),'@,@')):
                        if not c in vars[n]['check']:
                            vars[n]['check'].append(c)
                    check=None
            if dim and not vars[n].has_key('dimension'):
                vars[n]['dimension']=[]
                for d in rmbadname(map(string.strip,string.split(markoutercomma(dim),'@,@'))):
                    if d==':': d='*'
                    dl=string.split(markoutercomma(d,':'),'@:@')
                    if len(dl)==1 and not dl[0]=='*': dl = ['1',dl[0]]
                    if len(dl)==2:
                        d,v,di = getarrlen(dl,block['args'])
                        if d[:4] == '1 * ': d = d[4:]
                        if di and di[-4:] == '/(1)': di = di[:-4]
                        if v: savelindims[d] = v,di                        
                    vars[n]['dimension'].append(d)
        if vars[n].has_key('dimension'):
            vars[n]['dimension'].reverse()
            if isstringarray(vars[n]):
                if len(vars[n]['dimension'])==1:
                    if not vars[n].has_key('charselector'): vars[n]['charselector']={}
                    vars[n]['charselector']['*']='(%s)'%vars[n]['dimension'][0]
                    errmess('analyzevars: character rank-1 array "character %s(%s)" is considered as "character %s*(%s)"\n'%(n,vars[n]['dimension'][0],n,vars[n]['dimension'][0]))
                    del vars[n]['dimension']
                else:
                    if vars[n].has_key('charselector') and vars[n]['charselector'].has_key('*'):
                        del vars[n]['charselector']['*']
                    errmess('analyzevars: multi-D string arrays are not fully supported. "%s" = %s.\n'%(n,`vars[n]`))                
        if not vars[n].has_key('check') and block.has_key('args') and n in block['args']:
            flag=not vars[n].has_key('depend')
            if flag: vars[n]['depend']=[]
            vars[n]['check']=[]
            if vars[n].has_key('dimension'):
                #/----< no check
                #vars[n]['check'].append('rank(%s)==%s'%(n,len(vars[n]['dimension'])))
                i=-1; ni=len(vars[n]['dimension'])
                for d in vars[n]['dimension']:
                    ddeps=[] # dependecies of 'd'
                    ad=''
                    pd=''
                    origd = d
                    if not vars.has_key(d):
                        if savelindims.has_key(d):
                            pd,ad='(',savelindims[d][1]
                            d = savelindims[d][0]
                        else:
                            for r in block['args']:
                                if not vars.has_key(r): continue
                                if re.match(r'.*?\b'+r+r'\b',d):
                                    ddeps.append(r)
                    if vars.has_key(d):
                        if vars[d].has_key('attrspec'):
                            for aa in vars[d]['attrspec']:
                                if aa[:6]=='depend':
                                    ddeps=ddeps+string.split((string.strip(aa[6:]))[1:-1],',')
                        if vars[d].has_key('depend'):
                            ddeps=ddeps+vars[d]['depend']
                    i=i+1
                    if vars.has_key(d) and (not vars[d].has_key('depend')) and (not vars[d].has_key('=')) and (d not in vars[n]['depend']):
                        vars[d]['depend']=[n]
                        if ni>1:
                            vars[d]['=']='%sshape(%s,%s)%s'% (pd,n,i,ad)
                        else: vars[d]['=']='%slen(%s)%s'% (pd,n,ad)
                        #  /---< no check
                        if 1 and not vars[d].has_key('check'):
                            if ni>1: vars[d]['check']=['%sshape(%s,%i)%s==%s'%(pd,n,i,ad,d)]
                            else: vars[d]['check']=['%slen(%s)%s>=%s'%(pd,n,ad,d)]
                        if not vars[d].has_key('attrspec'): vars[d]['attrspec']=['optional']
                        if ('optional' not in vars[d]['attrspec']) and\
                           ('required' not in vars[d]['attrspec']):
                            vars[d]['attrspec'].append('optional')
                    elif d not in ['*',':']:
                        #/----< no check 
                        #if ni>1: vars[n]['check'].append('shape(%s,%i)==%s'%(n,i,d))
                        #else: vars[n]['check'].append('len(%s)>=%s'%(n,d))
                        if flag and vars.has_key(d):
                            if n not in ddeps:
                                vars[n]['depend'].append(d)
                        elif flag:
                            vars[n]['depend'] = vars[n]['depend'] + ddeps
            elif isstring(vars[n]):
                length='1'
                if vars[n].has_key('charselector'):
                    if vars[n]['charselector'].has_key('*'):
                        length=vars[n]['charselector']['*']
                        if length=='(:)':
                            vars[n]['charselector']['*']='(*)'
                            length='(*)'
                    elif vars[n]['charselector'].has_key('len'):
                        length=vars[n]['charselector']['len']
                        if length in ['(:)',':','*']:
                            vars[n]['charselector']['*']='(*)'
                            length='(*)'
                            del vars[n]['charselector']['len']
                #  /----< no check
                if 0 and (not length=='(*)'):
                    vars[n]['check'].append('slen(%s)>=%s'%(n,length))
                    if vars[n].has_key('='):
                        val=vars[n]['='][1:-1]
                        try:
                            if len(eval("'%s'"%val))<eval(length):
                                errmess('analyzevars: initial string %s="%s" has length smaller than required "%s".\n'%(n,val,length))
                        except:
                            errmess('analyzevars: failed to check (len(\'%s\')>=%s) for %s.\n'%(val,length,n))
                    
            if not vars[n]['check']: del vars[n]['check']
            if flag and not vars[n]['depend']: del vars[n]['depend']
        if vars[n].has_key('='):
            if not vars[n].has_key('attrspec'): vars[n]['attrspec']=[]
            if ('optional' not in vars[n]['attrspec']) and \
               ('required' not in vars[n]['attrspec']):
                vars[n]['attrspec'].append('optional')
            if not vars[n].has_key('depend'):
                vars[n]['depend']=[]
                for v in vars.keys():
                    if re.match(r'.*\b%s\b'%(v),vars[n]['=']): vars[n]['depend'].append(v)
                if not vars[n]['depend']: del vars[n]['depend']
    for n in vars.keys():
        if n==block['name']: # n is block name
            if vars[n].has_key('note'):
                block['note']=vars[n]['note']
            if block['block']=='function':
                if block.has_key('result') and vars.has_key(block['result']):
                    vars[n]=appenddecl(vars[n],vars[block['result']])
                if block.has_key('prefix'):
                    pr=block['prefix']; ispure=0; isrec=1
                    pr1=string.replace(pr,'pure','')
                    ispure=(not pr==pr1)
                    pr=string.replace(pr1,'recursive','')
                    isrec=(not pr==pr1)
                    m=typespattern[0].match(pr)
                    if m:
                        typespec,selector,attr,edecl=cracktypespec0(m.group('this'),m.group('after'))
                        kindselect,charselect,typename=cracktypespec(typespec,selector)
                        vars[n]['typespec']=typespec
                        if kindselect: vars[n]['kindselector']=kindselect
                        if charselect: vars[n]['charselector']=charselect
                        if typename: vars[n]['typename']=typename
                        if ispure: vars[n]=setattrspec(vars[n],'pure')
                        if isrec: vars[n]=setattrspec(vars[n],'recursive')
                    else:
                        outmess('analyzevars: prefix (%s) were not used\n'%`block['prefix']`)
    if not block['block'] in ['module','block data']:
        if block.has_key('commonvars'):
            neededvars=copy.copy(block['args']+block['commonvars'])
        else:
            neededvars=copy.copy(block['args'])
        if block['block']=='function':
            if block.has_key('result'):
                neededvars.append(block['result'])
            else:
                neededvars.append(block['name'])
        for n in vars.keys():
            if n not in neededvars:
                del vars[n]
    return vars

def analyzeargs(block):
    setmesstext(block)
    implicitrules,attrrules=buildimplicitrules(block)
    if not block.has_key('args'): block['args']=[]
    args=[]
    for a in block['args']:
        if not re.match(r'\A[a-z]+[\w$]*\Z',a): # `a` is an expression
            at=determineexprtype(a,block['vars'],implicitrules)
            na='e_'
            for c in a:
                if c not in string.lowercase+string.digits: c='_'
                na=na+c
            if na[-1]=='_': na=na+'e'
            else: na=na+'_e'
            a=na
            while block['vars'].has_key(a) or a in block['args']: a=a+'r'
            block['vars'][a]=at            
        args.append(a)
        if not block['vars'].has_key(a):
            block['vars'][a]={}
        if block.has_key('externals') and a in block['externals']+block['interfaced']:
            block['vars'][a]=setattrspec(block['vars'][a],'external')
    block['args']=args
    for b in block['body']:
        if b['name'] in args:
            if not block.has_key('externals'): block['externals']=[]
            if b['name'] not in block['externals']:
                block['externals'].append(b['name'])
    if block.has_key('result') and not block['vars'].has_key(block['result']):
        block['vars'][block['result']]={}
    return block
def determineexprtype(expr,vars,rules={}):
    if vars.has_key(expr): return vars[expr]
    expr=string.strip(expr)
    if re.match(r'\A\(.+?[,].+?\)\Z',expr):
        return {'typespec':'complex'}
    m=re.match(r'\A[+-]?\d+(_(P<name>[\w]+)|)?\Z',expr)
    if m:
        if m.groupdict().has_key('name') and m.group('name'):
            outmess('determineexprtype: selected kind types not supported (%s)\n'%`expr`)
        return {'typespec':'integer'}
    m=re.match(r'\A[+-]?[\d.]+[\d+-de.]*(_(P<name>[\w]+)|)?\Z',expr)
    if m:
        if m.group('name'):
            outmess('determineexprtype: selected kind types not supported (%s)\n'%`expr`)
        return {'typespec':'real'}
    for op in ['+','-','*','/']:
        for e in map(string.strip,string.split(markoutercomma(expr,comma=op),'@'+op+'@')):
            if vars.has_key(e): return vars[e]
    t={}
    if re.match(r'\A\(.*\)\Z',expr): # in parenthesis
        t=determineexprtype(expr[1:-1],vars,rules)
    else:
        m=re.match(r'\A(?P<name>\w+)\s*\(.*?\)\s*\Z',expr) # array element
        if m:
            rn=m.group('name')
            t=determineexprtype(m.group('name'),vars,rules)
            if t and t.has_key('attrspec'): del t['attrspec']
            if not t:
                if rules.has_key(rn[0]):
                    return rules[rn[0]]
    if not t:
        outmess('determineexprtype: could not determine expressions (%s) type.\n'%(`expr`))
    return t
######
def crack2fortrangen(block,tab='\n'):
    setmesstext(block)
    ret=''
    if type(block) is type([]):
        for g in block:
            ret=ret+crack2fortrangen(g,tab)
        return ret
    prefix=''
    name=''
    args=''
    blocktype=block['block']
    if blocktype=='program': return ''
    al=[]
    if block.has_key('name'): name=block['name']
    if block.has_key('args'):
        al=block['args']
        if block['block']=='function' or al:
            args='(%s)'%string.join(al,',')
    use=''
    if block.has_key('use'):
        use=use2fortran(block['use'],tab+tabchar)
    common=''
    if block.has_key('common'):
        common=common2fortran(block['common'],tab+tabchar)
    if name=='unknown_interface': name=''
    result=''
    if block.has_key('result'):
        result=' result (%s)'%block['result']
        al.append(block['result'])
    #if block.has_key('prefix'): prefix=block['prefix']+' '
    body=crack2fortrangen(block['body'],tab+tabchar)
    vars=vars2fortran(block,block['vars'],al,tab+tabchar)
    mess=''
    if block.has_key('from'):
        mess='! in %s'%block['from']
    ret='%s%s%s %s%s%s %s%s%s%s%s%send %s %s'%(tab,prefix,blocktype,name,args,result,mess,use,vars,common,body,tab,blocktype,name)
    return ret
def common2fortran(common,tab=''):
    ret=''
    for k in common.keys():
        ret='%s%scommon /%s/ %s'%(ret,tab,k,string.join(common[k],','))
    return ret
def use2fortran(use,tab=''):
    ret=''
    for m in use.keys():
        ret='%s%suse %s,'%(ret,tab,m)
        if use[m]=={}:
            if ret and ret[-1]==',': ret=ret[:-1]
            continue
        if use[m].has_key('only') and use[m]['only']:
            ret='%s,only:'%(ret)
        if use[m].has_key('map') and use[m]['map']:
            c=' '
            for k in use[m]['map'].keys():
                if k==use[m]['map'][k]:
                    ret='%s%s%s'%(ret,c,k); c=','
                else:
                    ret='%s%s%s=>%s'%(ret,c,k,use[m]['map'][k]); c=','
        if ret and ret[-1]==',': ret=ret[:-1]
    return ret
def vars2fortran(block,vars,args,tab=''):
    """
    TODO:
    public sub
    ...
    """
    setmesstext(block)
    ret=''
    nout=[]
    for a in args:
        if block['vars'].has_key(a): nout.append(a)
    if block.has_key('commonvars'):
        for a in block['commonvars']:
            if vars.has_key(a):
                if a not in nout: nout.append(a)
            else: errmess('vars2fortran: Confused?!: "%s" is not defined in vars.\n'%a)
    for a in vars.keys():
        if a not in nout: nout.append(a)
    for a in nout:
        if vars[a].has_key('depend'):
            for d in vars[a]['depend']:
                if vars.has_key(d) and vars[d].has_key('depend') and a in vars[d]['depend']:
                    errmess('vars2fortran: Warning: cross-dependence between variables "%s" and "%s"\n'%(a,d))
        if block.has_key('externals') and a in block['externals']:
            ret='%s%sexternal %s'%(ret,tab,a)
            if vars.has_key(a) and not vars[a].has_key('typespec'):
                continue
            cont=1
            for b in block['body']:
                if a==b['name'] and b['block']=='function': cont=0;break
            if cont: continue
        if not vars.has_key(a):
            show(vars)
            outmess('vars2fortran: No definition for argument "%s".\n'%a)
            continue
        if a==block['name'] and not block['block']=='function':
            continue
        if not vars[a].has_key('typespec'):
            if vars[a].has_key('attrspec') and 'external' in vars[a]['attrspec']:
                ret='%s%sexternal %s'%(ret,tab,a)
                continue
            show(vars[a])
            outmess('vars2fortran: No typespec for argument "%s".\n'%a)
            continue
        vardef=vars[a]['typespec']
        if vardef=='type' and vars[a].has_key('typename'):
            vardef='%s(%s)'%(vardef,vars[a]['typename'])
        selector={}
        if vars[a].has_key('kindselector'): selector=vars[a]['kindselector']
        elif vars[a].has_key('charselector'): selector=vars[a]['charselector']
        if selector.has_key('*'):
            if selector['*'] in ['*',':']:
                vardef='%s*(%s)'%(vardef,selector['*'])
            else:
                vardef='%s*%s'%(vardef,selector['*'])
        else:
            if selector.has_key('len'):
                vardef='%s(len=%s'%(vardef,selector['len'])
                if selector.has_key('kind'):
                    vardef='%s,kind=%s)'%(vardef,selector['kind'])
                else:
                    vardef='%s)'%(vardef)
            elif selector.has_key('kind'):
                vardef='%s(kind=%s)'%(vardef,selector['kind'])
        c=' '
        if vars[a].has_key('attrspec'):
            attr=[]
            for l in vars[a]['attrspec']:
                if l not in ['external']:
                    attr.append(l)
            if attr:
                vardef='%s %s'%(vardef,string.join(attr,','))
                c=','
        if vars[a].has_key('dimension'):
            vars[a]['dimension'].reverse()
        if vars[a].has_key('dimension'):
            vardef='%s%sdimension(%s)'%(vardef,c,string.join(vars[a]['dimension'],','))
            c=','
        if vars[a].has_key('intent'):
            vardef='%s%sintent(%s)'%(vardef,c,string.join(vars[a]['intent'],','))
            c=','
        if vars[a].has_key('check'):
            vardef='%s%scheck(%s)'%(vardef,c,string.join(vars[a]['check'],','))
            c=','
        if vars[a].has_key('depend'):
            vardef='%s%sdepend(%s)'%(vardef,c,string.join(vars[a]['depend'],','))
            c=','
        if vars[a].has_key('='):
            vardef='%s :: %s=%s'%(vardef,a,vars[a]['='])
        else:
            vardef='%s :: %s'%(vardef,a)
        ret='%s%s%s'%(ret,tab,vardef)
    return ret
######
skipfunctions=[]
skipfuncs=[]
onlyfuncs=[]
def crackfortran(files):
    global usermodules
    outmess('Reading fortran codes...\n',0)
    readfortrancode(files,crackline)
    outmess('Post-processing...\n',0)
    usermodules=[]
    postlist=postcrack(grouplist[0])
    return usermodules+postlist
def crack2fortran(block):
    global f2py_version
    pyf=crack2fortrangen(block)+'\n'
    header='!%f90 -*- f90 -*-'
    footer="""
! This file was auto-generated with f2py (version:%s).
! See http://cens.ioc.ee/projects/f2py2e/
"""%(f2py_version)
    return header+pyf+footer
if __name__ == "__main__":
    files=[]
    funcs=[]
    f=1;f2=0;f3=0
    showblocklist=0
    for l in sys.argv[1:]:
        if l=='': pass
        elif l[0]==':':
            f=0
        elif l=='-quite':
            quite=1
            verbose=0
        elif l=='-verbose':
            verbose=1
            quite=0
        elif l=='-fix':
            if strictf77:
                outmess('Use option -f90 before -fix if Fortran 90 code is in fix form.\n',0)
            skipemptyends=1
            sourcecodeform='fix'
        elif l=='-skipemptyends':
            skipemptyends=1
        elif l=='--ignore-contains':
            ignorecontains=1
        elif l=='-f77':
            strictf77=1
            sourcecodeform='fix'
        elif l=='-f90':
            strictf77=0
            sourcecodeform='free'
            skipemptyends=1
        elif l=='-h':
            f2=1
        elif l=='-show':
            showblocklist=1
        elif l=='-m':
            f3=1
        elif l[0]=='-':
            errmess('Unknown option %s\n'%`l`)
        elif f2:
            f2=0
            pyffilename=l
        elif f3:
            f3=0
            f77modulename=l
        elif f:
            try:
                open(l).close()
                files.append(l)
            except IOError,detail:
                errmess('IOError: %s\n'%str(detail))
        else:
            funcs.append(l)
    if not strictf77 and f77modulename and not skipemptyends:
        outmess("""\
  Warning: You have specifyied module name for non Fortran 77 code
  that should not need one (expect if you are scanning F90 code
  for non module blocks but then you should use flag -skipemptyends
  and also be sure that the files do not contain programs without program statement).
""",0)

    postlist=crackfortran(files,funcs)
    if pyffilename:
        outmess('Writing fortran code to file %s\n'%`pyffilename`,0)
        pyf=crack2fortran(postlist)
        f=open(pyffilename,'w')
        f.write(pyf)
        f.close()
    if showblocklist:
        show(postlist)
